import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_chat/src/repository/message.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:ulid/ulid.dart';
import 'package:ocg_user/ocg_user.dart';
import 'package:ocg_chat/src/models/message.dart';

import 'event.dart';
import 'state.dart';

class MessageBloc extends Bloc<MessageEvent, MessageState> {
  final MessageRepository messageRepository;
  final UserRepository userRepository;
  MessageBloc(this.messageRepository, this.userRepository)
      : super(MessageIdleState());

  @override
  Stream<MessageState> mapEventToState(MessageEvent event) async* {
    debugPrint("Event $event");
    if (event is MessageFetchEvent) {
      if (event.before == null) {
        yield MessageLoadingState();

        AppResponse<List<Message>> response = await messageRepository
            .getMessages(event.chatId, before: event.before);
        if (response.isSuccess)
          yield MessagedLoadedState(response.data);
        else {
          yield MessageErrorState(response.error);
        }
      } else {
        if (state is MessagedLoadedState &&
            (state as MessagedLoadedState).messages[0].chatId == event.chatId) {
          AppResponse<List<Message>> response = await messageRepository
              .getMessages(event.chatId, before: event.before);
          MessagedLoadedState loadedState = state;
          if (response.isSuccess) {
            List<Message> messages = loadedState.messages.toList();
            messages.addAll(response.data);
            messages.sort((a, b) => b.recordedAt.compareTo(a.recordedAt));
            yield MessagedLoadedState(messages);
          } else {
            yield MessageErrorState(response.error);
          }
        }
      }
    }
    if (event is MessageSendEvent) {
      Message message = event.message;

      AppResponse<String> loginResponse = await userRepository.getSelfId();
      if (loginResponse.isSuccess) {
        String senderId = loginResponse.data;
        String id = Ulid().toUuid();
        message = message.copyWith(
            id: id,
            randId: id,
            senderId: senderId,
            recordedAt: DateTime.now(),
            isSelf: true,
            state: MessageNetworkState.progress);
        yield MessagedLoadedState(addMessage(message));
        AppResponse<MessageResponse> response =
            await messageRepository.sendMessage(message);
        if (response.isSuccess) {
          if (message.isSelf) {
            AppResponse<User> userResponse = await userRepository.getSelf();
            if (userResponse.isSuccess) {
              message = message.copyWith(sender: userResponse.data);
            }
          }
          if (state is MessagedLoadedState) {
            message = message.copyWith(
                id: response.data.msgId,
                randId: response.data.randId,
                state: MessageNetworkState.sent);
            // TODO Add function
            yield MessagedLoadedState(addMessage(message));
            // oldMessages.insert(0, message);
            // oldMessages.sort((a, b) => a.recordedAt.compareTo(b.recordedAt));
            // yield MessagedLoadedState(oldMessages);
          }
        }
      }
    }
    if (event is MessageRetryEvent) {
      // await messageRepository.retryMessage(event.message);
    }
    if (event is MessageReceiveEvent) {
      AppResponse<String> loginResponse = await userRepository.getSelfId();

      if (loginResponse.isSuccess &&
          loginResponse.data != event.message.senderId) {
        Message message = event.message;

        AppResponse<User> userResponse =
            await userRepository.getUser(event.message.senderId);
        message = message.copyWith(
          isSelf: message.senderId == loginResponse.data,
        );
        if (userResponse.isSuccess) {
          message = message.copyWith(sender: userResponse.data);
        }
        List<Message> messages = addMessage(message);
        yield MessagedLoadedState(messages);
      }
    }
    if (event is MessageSeenEvent) {
      if (state is MessagedLoadedState) {
        MessagedLoadedState loadedState = state;
        List<Message> messages = List<Message>.from(loadedState.messages);
        int index = messages.indexWhere((val) => val.id == event.msgId);
        if (index != -1) {
          Message message = messages[index];
          List<Message> readMessages = [];
          messages.forEach((val) {
            if (val.recordedAt.compareTo(message.recordedAt) <= 0 &&
                !(val.isSeen ?? false)) {
              readMessages.add(val.copyWith());
            }
          });
          readMessages.forEach((val) {
            val = val.copyWith(isSeen: true);
            messages.removeWhere((temp) => temp.id == val.id);
            messages.add(val);
          });
          messages.removeWhere((temp) => message.id == temp.id);
          Message newMessage = message.copyWith(isSeen: true);
          messages.add(newMessage);
          messages.sort((a, b) => b.recordedAt.compareTo(a.recordedAt));

          yield MessagedLoadedState(messages);
        }
      }
    }
    if (event is MessageReadEvent) {
      Message message = event.message;
      if (!(message.isSeen ?? false) && !(message.isSelf ?? false))
        await messageRepository.readChat(message.chatId, message.id);
    }
  }

  List<Message> addMessage(Message message) {
    List<Message> messages = [];
    if (state is MessagedLoadedState) {
      messages = List<Message>.from((state as MessagedLoadedState).messages);
      if (messages
              .where((element) =>
                  element.randId == message.randId && element.randId != null)
              .length >
          0) {
        messages.removeWhere((el) => el.randId == message.randId);
      }
      if (messages.where((element) => element.id == message.id).length > 0) {
        messages.removeWhere((el) => el.id == message.id);
      }
    }
    messages.add(message);
    messages.sort((a, b) => b.recordedAt.compareTo(a.recordedAt));
    return messages;
  }
}
